import click
import sys
import os
import json
from pathlib import Path
from anc.cli.util import click_group, console
from .util import ConfigManager
import uuid
import datetime
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.console import Console
from rich import box
from .operators.eval_operator import trigger_eval_job, display_evaluation_status

@click_group()
def eval():
    pass

@eval.command()
@click.argument('model_name', required=True, type=click.Choice(['ds_v2', 'ds_v2_lite', 'qwen25_7b', 'qwen25_14b', 'qwen25_32b']))
@click.option(
    '--dataset_paths',
    type=str,
    required=True,
    help='the eval dataset path of list/single of dataset'
)
@click.option(
    '--ckpt_paths',
    type=str,
    required=True,
    help='the eval ckpt path of list of ckpt'
)
@click.option(
    '--tp',
    type=int,
    required=False,
    default=8,
    help='the evaltensor parallel size'
)
@click.option(
    '--pp',
    type=int,
    required=False,
    default=1,
    help='the eval pipeline parallel size'
)
@click.option(
    '--ep',
    type=int,
    required=False,
    default=1,
    help='the eval expert parallel size'
)
@click.option(
    '--seq_len',
    type=int,
    required=True,
    help='the eval sequence length'
)
@click.option(
    '--batch_size',
    type=int,
    required=True,
    help='the eval batch size'
)
@click.option(
    '--tokenizer_path',
    type=str,
    required=False,
    default="/mnt/project/llm/ckpt/tokenizer/ocean_deepseek_v2",
    help='the project name'
)
@click.option(
    '--project_name',
    type=str,
    required=False,
    default="my_test",
    help='the project name'
)
@click.pass_context
def model(ctx, model_name, ckpt_paths, dataset_paths, tp, pp, ep, seq_len, batch_size, tokenizer_path, project_name):
    """command like: anc eval ds_v2 --ckpt_paths  --ckpt_paths ckpt_paths --dataset_paths dataset_paths"""
    print("args:", model_name, ckpt_paths, dataset_paths)
    if  not ckpt_paths.startswith("/mnt/project") and not ckpt_paths.startswith("/mnt/share"):
        print("ckpt path is invalid, must be start with /mnt/project or /mnt/share")
        sys.exit(0)
    
    if  not dataset_paths.startswith("/mnt/project") and not dataset_paths.startswith("/mnt/share"):
        print("dataset path is invalid, must be start with /mnt/project or /mnt/share")
        sys.exit(0)
        
    # the recipt of the eval
    run_id = str(uuid.uuid4())
    
    # Create a rich Table for configuration details
    config_table = Table(box=box.ROUNDED, expand=True, show_header=False, border_style="blue")
    config_table.add_column("Parameter", style="cyan", no_wrap=True)
    config_table.add_column("Value", style="green")
    
    # Add basic info
    config_table.add_row("Model Name", model_name)
    config_table.add_row("Checkpoint Path", ckpt_paths)
    config_table.add_row("Dataset Path", dataset_paths)
    config_table.add_row("Tokenizer Path", tokenizer_path)
    config_table.add_row("Project Name", project_name)
    # Add parallelism config
    config_table.add_section()
    config_table.add_row("Tensor Parallel", str(tp))
    config_table.add_row("Pipeline Parallel", str(pp))
    config_table.add_row("Expert Parallel", str(ep))
    
    # Add run parameters
    config_table.add_section()
    config_table.add_row("Batch Size", str(batch_size))
    config_table.add_row("Sequence Length", str(seq_len))

    # Create title with run ID
    title = Text(f"✨ EVALUATION RECEIPT [ID: {run_id}] ✨", style="bold magenta")
    
    # Wrap the table in a panel
    panel = Panel(
        config_table,
        title=title,
        subtitle="Your evaluation receipt!",
        border_style="blue",
        expand=False
    )
    
    # Print the panel using the console
    console.print("\n")
    console.print(panel)
    console.print("\n")
    # TODO: Implement need the user confirm the evaluation receipt
    user_confirm = input("Do you want to start the evaluation? (y/n): ")
    if user_confirm.lower() != 'y':
        print("Evaluation cancelled.")
        sys.exit(0)
    
    eval_ckpt_paths_list = []
    if os.path.isdir(ckpt_paths):
        for ckpt_path in os.listdir(ckpt_paths):
            eval_ckpt_paths_list.append(os.path.join(ckpt_paths, ckpt_path))
    else:
        raise ValueError("ckpt_paths is not a directory")
    
    if not os.path.isdir(dataset_paths):
        raise ValueError("dataset_paths is not a directory")
    trigger_eval_job(run_id, model_name, project_name, eval_ckpt_paths_list, dataset_paths, tp, pp, ep, seq_len, batch_size, tokenizer_path)
    
    
    

@eval.command(name='status')
@click.argument('eval_id', required=True, type=str)
def status(eval_id):
   """check eval job: anc eval status xxx """
   display_evaluation_status(eval_id)

def add_command(cli_group):
    cli_group.add_command(eval)


from anc.data.anc_dataloader import AncDataLoader
from anc.data.processors.llm_processor import PretrainProcessor
from anc.data.anc_composer import SeqSplitConfig
from anc.data.parquet_dataset import ParquetDataset
from transformers import AutoTokenizer
import concurrent.futures
from functools import partial
import json
import os

def get_token_count(ds_path, tokenizer_path):
    try:
        ds = ParquetDataset(ds_path)
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
        tokens = 0
        for item in ds:
            tokens += len(tokenizer.encode(item['content']))
        print(f"Processed {ds_path}: {tokens} tokens")
        return tokens
    except Exception as e:
        print(f"Error processing {ds_path}: {e}")
        return 0

def get_dataset_paths(path):
    dataset_paths = []
    if os.path.isdir(path) and 'eval' in os.listdir(path):
        dataset_paths.append(path)
        return dataset_paths

    # Add all subdirectories as dataset paths
    subdirs = [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
    for subdir in subdirs:
        subdir_path = os.path.join(path, subdir)
        dataset_paths.extend(get_dataset_paths(subdir_path))

    return dataset_paths


def count_eval_tokens(dataset_dir, tokenizer_path, max_workers=4):
    eval_dir = os.path.join(dataset_dir, "eval")

    if not os.path.exists(eval_dir):
        print(f"Eval directory not found in {dataset_dir}")
        return 0

    parquet_files = []
    for file in os.listdir(eval_dir):
        if file.endswith(".snappy.parquet"):
            parquet_files.append(os.path.join(eval_dir, file))

    if not parquet_files:
        print(f"No parquet files found in {eval_dir}")
        return 0

    print(f"Found {len(parquet_files)} parquet files in {eval_dir}")

    total_tokens = 0
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_path = {
            executor.submit(get_token_count, path, tokenizer_path): path 
            for path in parquet_files
        }

        for future in concurrent.futures.as_completed(future_to_path):
            path = future_to_path[future]
            try:
                token_count = future.result()
                total_tokens += token_count
                print(f"Processed {path}: {token_count} tokens")
            except Exception as e:
                print(f"Error processing {path}: {e}")

    return total_tokens


# Process all eval dataset token count and dump to json file [offline use only].
#   a. If the key of tokenizer in json file, read the token count from the json file.
#   b. If the key of tokenizer not in json file, calculate the token count and dump to the json file.
#   c. For the eval dataset, the token count is the sum of all the token count of the parquet files in the eval folder. 
def process_all_datasets(dataset_list, tokenizer_path):
    results = {}

    for dataset_dir in dataset_list:
        dataset_name = os.path.basename(dataset_dir)
        print(f"\nProcessing dataset: {dataset_name}")

        eval_dir = os.path.join(dataset_dir, "eval")
        if os.path.exists(eval_dir):
            json_path = os.path.join(eval_dir, "token_count.json")
            if os.path.exists(json_path):
                try:
                    with open(json_path, 'r') as f:
                        existing_data = json.load(f)
                    if tokenizer_path in existing_data:
                        token_count = existing_data[tokenizer_path]
                        print(f"Using existing token count for {dataset_name}: {token_count}")
                        results[dataset_dir] = token_count
                        continue
                except Exception as e:
                    print(f"Error reading existing JSON: {e}")

        token_count = count_eval_tokens(dataset_dir, tokenizer_path)
        results[dataset_dir] = token_count

        if os.path.exists(eval_dir):
            json_path = os.path.join(eval_dir, "token_count.json")

            existing_data = {}
            if os.path.exists(json_path):
                try:
                    with open(json_path, 'r') as f:
                        existing_data = json.load(f)
                    print(f"Found existing token count file at {json_path}")
                except json.JSONDecodeError:
                    print(f"Waring: reading existing JSON file {json_path}, will create new one")

            existing_data[tokenizer_path] = token_count
            
            with open(json_path, 'w') as f:
                json.dump(existing_data, f, indent=2)
            print(f"Updated token count at {json_path}")

    print("\n===== Token Count Summary =====")
    for dataset_dir, count in results.items():
        print(f"{os.path.basename(dataset_dir)}: {count} tokens")

    return results


def gen_eval_data_cache_meta(dataset_dir, tokenizer_path):
    dataset_paths = get_dataset_paths(file_dir)
    process_all_datasets(dataset_paths, tokenizer_path)


# Online use.
def get_num_iterations(ds_path, tokenizer_path, global_batch_size, seq_len):
    num_workers = min(16, len(ds_path))
    total_token_count = 0
    
    with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
        future_to_path = {
            executor.submit(get_token_count, path, tokenizer_path): path 
            for path in ds_path
        }

        for future in concurrent.futures.as_completed(future_to_path):
            path = future_to_path[future]
            try:
                token_count = future.result()
                total_token_count += token_count
                print(f"Processed eval dataset {path}: {token_count} tokens")
            except Exception as e:
                print(f"Error processing eval dataset {path}: {e}")
    return total_token_count // (global_batch_size * seq_len)


# Online use.
def get_num_iterations_from_cache(file_dir, tokenizer_path):
    token_count = 0
    cached_dataset = 0
    dataset_paths = get_dataset_paths(file_dir)
    for dataset_path in dataset_paths:
        if os.path.exists(os.path.join(dataset_path, "eval", "token_count.json")):
            with open(os.path.join(dataset_path, "eval", "token_count.json"), 'r') as f:
                data = json.load(f)
                if tokenizer_path in data:
                    token_count += int(data.get(tokenizer_path, 0))
                    cached_dataset += 1
    return token_count, cached_dataset, len(dataset_paths)


# For testing.
if __name__ == "__main__":
    file_dir = "/mnt/project/llm/data/pretrain/D0_8/"
    #file_dir = "/mnt/project/llm/users/xug/D0_8/rentry_web_stories"
    
    if not os.path.exists(file_dir):
        raise FileNotFoundError(f"Dataset directory {file_dir} does not exist")
    
    dataset_paths = get_dataset_paths(file_dir)
    print(len(dataset_paths), dataset_paths)

    #process_all_datasets(dataset_paths, "/mnt/project/llm/ckpt/tokenizer/ocean_deepseek_v2")
    count, cached_dataset, total_dataset = get_num_iterations_from_cache(file_dir, "/mnt/project/llm/ckpt/tokenizer/ocean_deepseek_v2")
    print(count, cached_dataset, total_dataset, cached_dataset / total_dataset)
    #paths = ["/mnt/project/llm/pretrain_datasets/pretrain_dataset_v0_7/thePile-BooksCorpus2/eval/part-00000-22d7774e-fc6c-4076-ab62-64d6249f017f-c000.snappy.parquet"]


from anc.api.connection import Connection
from requests.exceptions import RequestException
import os
import sys
import json
from rich.console import Console
from rich.table import Table, box
from rich.text import Text

MM_BASE_URL = "http://model-management-service.infra.svc.cluster.local:5000"


def trigger_eval_job(
    run_id: str,
    model_name: str,
    project_name: str,
    ckpt_paths_list: list[str],
    dataset_paths: list[str],
    tp: int,
    pp: int,
    ep: int,
    seq_len: int,
    batch_size: int,
    tokenizer_path: str,
) -> bool:
    cluster = os.environ.get("MLP_CLUSTER", "il2")
    project = os.environ.get("MLP_PROJECT", "llm")
    
    data = {
        "evaluation_id": run_id,
        "modality": "nlp",
        "model_name": model_name,
        "project_name": project_name,
        "eval_ckpt_path_list": ckpt_paths_list,
        "eval_dataset_path": dataset_paths,
        "project": project,
        "cluster": cluster,
        "eval_tp": tp,
        "eval_pp": pp,
        "eval_ep": ep,
        "eval_seqlen": seq_len,
        "eval_batch_size": batch_size,
        "eval_tokenizer_path": tokenizer_path,
        "status": "start",
    }
    
    try:
        conn = Connection(url=MM_BASE_URL)
        response = conn.post("/evaluations", json=data)

        # Check if the status code is in the 2xx range
        if 200 <= response.status_code < 300:
            response_data = response.json()
            evaluation_id = response_data.get('evaluation_id')
            if evaluation_id:
                print(f"Evaluation task added successfully. Your Eval ID is: \033[92m{evaluation_id}\033[0m")
                print(f"You can check the status of your evaluation using: \033[96manc eval status {evaluation_id}\033[0m")
                print(f"All historical results can be viewed at: \033[94mhttp://model.anuttacon.ai/models/467e151d-a52a-47f9-8791-db9c776635db/evaluations\033[0m")
            else:
                print("Evaluation failed, didn't get the evaluation id")
        else:
            #print(f"Error: Server responded with status code {response.status_code}")
            print(f"{response.text}")

    except RequestException as e:
        print(f"Sorry, you can't add dataset out of clusters, please use it in a notebook")
    except json.JSONDecodeError:
        print("Sorry: received invalid JSON response from server")
    except KeyboardInterrupt:
        print(f"Operation interrupted.")
        sys.exit(0)
    except Exception as e:
        print(f"Sorry, your command run failed, you can try again or reach out infra team")


def display_evaluation_status(evaluation_id: str):
    conn = Connection(url=MM_BASE_URL)
    response = conn.get(f"/evaluations/{evaluation_id}")
    
    if response.status_code == 200:
        data = response.json()
        
        # Create a Rich console instance
        console = Console(width=200)  # Set wider console width
        
        # Display basic evaluation information
        eval_info = Table(title=f"Evaluation ID: {evaluation_id}", box=box.ROUNDED)
        eval_info.add_column("Parameter", style="cyan")
        eval_info.add_column("Value", style="green")
        
        # Add some key evaluation parameters
        eval_info.add_row("Model Name", data.get('model_name') or 'N/A')
        eval_info.add_row("Project", data.get('project') or 'N/A')
        eval_info.add_row("Submitted At", data.get('submitted_at') or 'N/A')
        
        console.print(eval_info)
        console.print()
        
        # Parse and display the evaluation_results_info
        if data.get('evaluation_results_info'):
            try:
                results_info = json.loads(data['evaluation_results_info'])
                
                # Create table for evaluation results with expanded width
                results_table = Table(title="Evaluation Results", box=box.ROUNDED, show_lines=True)
                results_table.add_column("Checkpoint", style="magenta", width=50, no_wrap=True)
                results_table.add_column("Dataset", style="blue", width=25, no_wrap=True)
                results_table.add_column("Endpoint URL", style="yellow", no_wrap=True)
                
                # Add rows for each checkpoint and dataset combination
                for ckpt_path, dataset_list in results_info.items():
                    # Get basename for the checkpoint
                    ckpt_basename = os.path.basename(ckpt_path)
                    
                    # Handle the case where each checkpoint has multiple datasets
                    for dataset_info in dataset_list:
                        if len(dataset_info) >= 4:
                            # Extract dataset info
                            dataset_path = dataset_info[0]
                            endpoint_url = dataset_info[1]
                            job_id = dataset_info[2]
                            status = dataset_info[3]
                            
                            # Get basename for dataset
                            dataset_basename = os.path.basename(dataset_path)
                            
                            results_table.add_row(
                                ckpt_basename,
                                dataset_basename,
                                endpoint_url
                            )
                
                # Ensure the table doesn't truncate content
                console.print(results_table)
            except json.JSONDecodeError:
                console.print(f"[red]Error parsing evaluation results info: {data['evaluation_results_info']}[/red]")
        else:
            console.print("[yellow]No evaluation results information available.[/yellow]")
    else:
        console.print(f"[red]Error retrieving evaluation status: {response.text}[/red]")
